# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (C) 2012 Coop. Trab. Moldeo Interactive Ltda.
# http://business.moldeo.coop
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################
import os, time
from M2Crypto import BIO, Rand, SMIME, EVP, RSA, X509, ASN1
from osv import fields, osv, orm
from tools.translate import _

class certificate(osv.osv):
    def _get_status(self, cr, uid, ids, field_name, arg, context=None):
        r = {}
        for cer in self.browse(cr, uid, ids):
            csr = cer.csr and cer.csr.strip()
            crt = cer.crt and cer.crt.strip()
            if not csr and not crt:
                r[cer.id] = 'Empty'
            elif csr and not crt:
                try:
                    req = cer.get_request()[cer.id]
                    pkey = req.get_pubkey()
                    if req.verify(pkey):
                        r[cer.id] = 'Valid Request'
                    else:
                        r[cer.id] = 'Invalid Request'
                except:
                    r[cer.id] = 'Invalid Request'
            elif csr and crt:
                pk = EVP.PKey()
                pk.assign_rsa(cer.pairkey_id.as_rsa()[cer.pairkey_id.id])
                try:
                    crt = cer.get_certificate()[cer.id]
                    r[cer.id] = 'Valid' if crt.verify(pk) else 'Invalid'
                except:
                    r[cer.id] = 'Invalid'
            else:
                r[cer.id] = 'Invalid'
        return r

    def _get_data(self, cr, uid, ids, field_name, arg, context=None):
        ret = {}
        for item in self.browse(cr, uid, ids, context=context):
            fields = ('req_subject','req_pubkey','crt_subject','crt_pubkey','crt_issuer','crt_fingerprint','is_ca')
            ret[item.id] = dict.fromkeys(fields, False)
            if item.csr:
                try:
                    req = X509.load_request_string(item.csr.encode('ascii'))
                    ret[item.id]  = {'req_subject': req.get_subject().as_text(),
                            'req_pubkey': req.get_pubkey().get_rsa().as_pem(),}
                except:
                    pass
            if item.crt:
                try:
                    crt = X509.load_cert_string(item.crt.encode('ascii'))
                    ret[item.id].update( {'crt_subject': crt.get_subject().as_text(),
                            'crt_pubkey': crt.get_pubkey().get_rsa().as_pem(),
                            'crt_issuer': crt.get_issuer().as_text(),
                            'crt_fingerprint': crt.get_fingerprint(),
                            'is_ca': crt.check_ca(),} )
                except:
                    pass
        return ret
    
    _name = "crypto.certificate"
    _columns = {
        'name': fields.char('Name', size=256),
        'csr': fields.text('Request Certificate',
                           readonly=True, states={'draft': [('readonly',False)]}, 
                           help='Certificate Request in PEM format.'),
        'crt': fields.text('Certificate',
                           readonly=True, states={'draft': [('readonly',False)], 'waiting':[('readonly',False)]}, 
                           help='Certificate in PEM format.'),
        'pairkey_id': fields.many2one('crypto.pairkey', 'Key pair'),
        'req_subject' :fields.function(_get_data, type='char', size=200, string='Req Subject', method=True, multi='data',),
        'req_pubkey' :fields.function(_get_data, type='text', string='Req Pubkey', method=True,multi='data',),
        'crt_subject' :fields.function(_get_data, type='char', size=200, string='Crt Subject', method=True, multi='data',),
        'crt_pubkey' :fields.function(_get_data, type='text', string='Crt Pubkey', method=True,multi='data',),
        'crt_issuer' :fields.function(_get_data, type='char', size=200, string='Crt Issuer', method=True, multi='data',),
        'crt_fingerprint' :fields.function(_get_data, type='text', string='Crt Fingerprint', method=True,multi='data',),
        'is_ca' : fields.function(_get_data, type='boolean', string='Is CA', method=True, multi='data',),
        'status': fields.function(_get_status, method=True, string='Status', type='char',
                                 help='Certificate Status'),
        'state': fields.selection([
                ('draft', 'Draft'),
                ('waiting', 'Waiting'),
                ('confirmed', 'Confirmed'),
                ('cancel', 'Cancelled'),
            ], 'State', select=True, readonly=True,
            help='* The \'Draft\' state is used when a user is creating a new pair key. Warning: everybody can see the key.\
            \n* The \'Waiting\' state is used when a request has send to Certificate Authority and is waiting for response.\
            \n* The \'Confirmed\' state is used when a certificate is valid.\
            \n* The \'Canceled\' state is used when the key is not more used. You cant use this key again.'),
    }
    _defaults = {
        'state': 'draft',
    }

    def action_validate(self, cr, uid, ids, context=None):
        if context is None:
            context = {}
        certs = self.read(cr, uid, ids, ['status', 'state'], context=context)
        sids = {'confirmed': [], 'waiting': [], 'invalid': [],}
        for cert in certs:
            status = cert['status']
            state = cert['state']
            if status == 'Valid Request' and state == 'draft':
                sids['waiting'].append(cert['id'])
            elif status == 'Valid' and state == 'waiting':
                sids['confirmed'].append(cert['id'])
        for state,ids in sids.items():
            self.write(cr, uid, ids, {'state': state,}, context=context)
        return True

    def action_cancel(self, cr, uid, ids, context=None):
        if context is None:
            context = {}
        self.write(cr, uid, ids, {'state': 'cancel'}, context=context)
        return True

    def get_request(self, cr, uid, ids, context=None):
        """
        Return Request object.
        """
        r = {}
        for cert in self.browse(cr, uid, ids):
            if cert.csr:
                r[cert.id] = X509.load_request_string(cert.csr.encode('ascii'))
        return r

    def get_certificate(self, cr, uid, ids, context=None):
        """
        Return Certificate object.
        """
        r = {}
        for cert in self.browse(cr, uid, ids):
            if cert.csr:
                r[cert.id] = X509.load_cert_string(cert.crt.encode('ascii'))
        return r

    def generate_certificate(self, cr, uid, ids, issuer,
                             ext=None, serial_number=1, version=2,
                             date_begin=None,
                             date_end=None,
                             expiration=365,
                             context=None):
        """
        Generate certificate
        """
        for item in self.browse(cr, uid, ids):
            if item.status == 'Valid Request':
                # Get request data
                pk = item.pairkey_id.as_pkey()[item.pairkey_id.id]
                req = item.get_request()[item.id]
                sub = req.get_subject()
                pkey = req.get_pubkey()
                # Building certificate
                cert = X509.X509()
                cert.set_serial_number(serial_number)
                cert.set_version(version)
                cert.set_subject(sub)

                now = ASN1.ASN1_UTCTIME()
                if date_begin is None:
                    t = long(time.time()) + time.timezone
                    now.set_time(t)
                else:
                    now.set_datetime(date_begin)

                nowPlusYear = ASN1.ASN1_UTCTIME()
                if date_end is None:
                    nowPlusYear.set_time(t + 60 * 60 * 24 * expiration)
                else:
                    nowPlusYear.set_datetime(date_end)

                cert.set_not_before(now)
                cert.set_not_after(nowPlusYear)
                cert.set_issuer(issuer)
                cert.set_pubkey(pkey)
                cert.set_pubkey(cert.get_pubkey())
                if ext:
                    cert.add_ext(ext)
                cert.sign(pk, 'sha1')
                w = { 'crt': cert.as_pem() }
                self.write(cr, uid, item.id, w)

    def smime(self, cr, uid, ids, message, context=None):
        """
        Sign message in SMIME format.
        """
        r = {}
        for cert in self.browse(cr, uid, ids):
            if cert.status == 'Valid':
                smime = SMIME.SMIME()
                ks = BIO.MemoryBuffer(cert.pairkey_id.key.encode('ascii'))
                cs = BIO.MemoryBuffer(cert.crt.encode('ascii'))
                bf = BIO.MemoryBuffer(str(message))
                out = BIO.MemoryBuffer()
                try:
                    smime.load_key_bio(ks, cs)
                except EVP.EVPError:
                    raise osv.except_osv(_('Error in Key and Certificate strings !'), _('Please check if private key and certificate are in ASCII PEM format.'))
                sbf = smime.sign(bf)
                smime.write(out, sbf)
                r[cert.id] = out.read()
            else:
                raise osv.except_osv(_('This certificate is not ready to sign any message !'), _('Please set a certificate to continue. You must send your certification request to a authoritative certificator to get one, or execute a self sign certification'))
        return r

certificate()

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:
